# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import itertools
import warnings

import numpy as np
import pandas as pd
from tqdm import tqdm

from probatus.binning import AgglomerativeBucketer, QuantileBucketer, SimpleBucketer
from probatus.stat_tests import ad, es, ks, psi, sw
from probatus.utils.arrayfuncs import check_numeric_dtypes


class DistributionStatistics:
    """
    Wrapper that applies a statistical test to compare two distributions.

    Details on the available tests can be found [here](/probatus/api/stat_tests.html#available-tests).

    For some tests, default data binning strategies are also provided.

    Example:
    ```python
    import numpy as np
    import pandas as pd
    from probatus.stat_tests import DistributionStatistics

    d1 = np.histogram(np.random.normal(size=1000), 10)[0]
    d2 = np.histogram(np.random.normal(size=1000), 10)[0]
    myTest = DistributionStatistics('KS', bin_count=10)
    test_statistic, p_value = myTest.compute(d1, d2, verbose=True)
    ```
    """

    binning_strategy_dict = {
        "simplebucketer": SimpleBucketer,
        "agglomerativebucketer": AgglomerativeBucketer,
        "quantilebucketer": QuantileBucketer,
        None: None,
    }
    statistical_test_dict = {
        "ES": {
            "func": es,
            "name": "Epps-Singleton",
            "default_binning": None,
        },
        "KS": {
            "func": ks,
            "name": "Kolmogorov-Smirnov",
            "default_binning": None,
        },
        "AD": {
            "func": ad,
            "name": "Anderson-Darling TS",
            "default_binning": None,
        },
        "SW": {
            "func": sw,
            "name": "Shapiro-Wilk based difference",
            "default_binning": None,
        },
        "PSI": {
            "func": psi,
            "name": "Population Stability Index",
            "default_binning": "quantilebucketer",
        },
    }

    def __init__(self, statistical_test, binning_strategy="default", bin_count=10):
        """
        Initializes the class.

        Args:
            statistical_test (str): Statistical test to apply. Available tests:

                - `'ES'`: Epps-Singleton
                - `'KS'`: Kolmogorov-Smirnov
                - `'PSI'`: Population Stability Index
                - `'SW'`: Shapiro-Wilk
                - `'AD'`: Anderson-Darling

                Details on the available tests can be found [here](/probatus/api/stat_tests.html#available-tests)

            binning_strategy (string, optional):
                Binning strategy to apply, binning strategies implemented:

                - `'simplebucketer'`: equally spaced bins,
                - `'agglomerativebucketer'`: binning by applying the Scikit-learn implementation of Agglomerative
                    Clustering,
                - `'quantilebucketer'`: bins with equal number of elements,
                - `'default'`: applies a default binning for a given stats_test. For all tests apart from PSI, no
                    binning (None) is used. For PSI by default quantilebucketer is used,
                - `None`: no binning is applied. The test is computed based on original distribution.

            bin_count (int, optional): In case binning_strategy is not None, specify the number of bins to be used by
                the binning strategy. By default 10 bins are used.
        """
        self.statistical_test = statistical_test.upper()
        self.binning_strategy = binning_strategy
        self.bin_count = bin_count
        self.fitted = False

        # Initialize the statistical test
        if self.statistical_test not in self.statistical_test_dict:
            raise NotImplementedError(f"The statistical test should be one of {self.statistical_test_dict.keys()}")
        else:
            self.statistical_test_name = self.statistical_test_dict[self.statistical_test]["name"]
            self._statistical_test_function = self.statistical_test_dict[self.statistical_test]["func"]

        # Initialize the binning strategy
        if self.binning_strategy:
            self.binning_strategy = self.binning_strategy.lower()
            if self.binning_strategy == "default":
                self.binning_strategy = self.statistical_test_dict[self.statistical_test]["default_binning"]
            if self.binning_strategy not in self.binning_strategy_dict:
                raise NotImplementedError(
                    f"The binning strategy should be one of {list(self.binning_strategy_dict.keys())}"
                )
            else:
                binner = self.binning_strategy_dict[self.binning_strategy]
                if binner is not None:
                    self.binner = binner(bin_count=self.bin_count)

    def __repr__(self):
        """
        String representation.
        """
        repr_ = f"DistributionStatistics object\n\tstatistical_test: {self.statistical_test}"
        if self.binning_strategy:
            repr_ += f"\n\tbinning_strategy: {self.binning_strategy}\n\tbin_count: {self.bin_count}"
        else:
            repr_ += "\n\tNo binning applied"
        if self.fitted:
            repr_ += f"\nResults\n\tvalue {self.statistical_test}-statistic: {self.statistic}"
        if hasattr(self, "p_value"):
            repr_ += f"\n\tp-value: {self.p_value}"
        return repr_

    def compute(self, d1, d2, verbose=False):
        """
        Apply the statistical test and compute statistic value and p-value.

        Args:
            d1 (np.array or pandas.DataFrame):
                distribution 1.

            d2 (np.array or pandas.DataFrame):
                distribution 2.

            verbose (bool, optional):
                Flag indicating whether prints should be shown.

        Returns:
            float: Statistic value
            float: p_value. For PSI test, only the statistic value is returned
        """
        check_numeric_dtypes(d1)
        check_numeric_dtypes(d2)

        # Bin the data
        if self.binning_strategy:
            self.binner.fit(d1)
            d1_preprocessed = self.binner.compute(d1)
            d2_preprocessed = self.binner.compute(d2)
        else:
            d1_preprocessed, d2_preprocessed = d1, d2

        # Perform the statistical test
        res = self._statistical_test_function(d1_preprocessed, d2_preprocessed, verbose=verbose)
        self.fitted = True

        # Check form of results and return
        if type(res) == tuple:
            self.statistic, self.p_value = res
            return self.statistic, self.p_value
        else:
            self.statistic = res
            return self.statistic


class AutoDist:
    """Apply stat tests and binning strategies.

    Class to automatically apply all implemented statistical distribution tests and binning strategies
    to (a selection of) features in two dataframes.

    Details on the available tests can be found [here](/probatus/api/stat_tests.html#available-tests).

    Example:
    ```python
    import numpy as np
    import pandas as pd
    from probatus.stat_tests import AutoDist

    df1 = pd.DataFrame(np.random.normal(size=(1000, 2)), columns=['feat_0', 'feat_1'])
    df2 = pd.DataFrame(np.random.normal(size=(1000, 2)), columns=['feat_0', 'feat_1'])
    myAutoDist = AutoDist(statistical_tests=["KS", "PSI"], binning_strategies='simplebucketer', bin_count=10)
    myAutoDist.compute(df1, df2, column_names=df1.columns)
    ```

    <img src="../img/autodist.png" width="700" />
    """

    def __init__(self, statistical_tests="all", binning_strategies="default", bin_count=10):
        """
        Initializes the class.

        Args:
            statistical_tests (str or list of str, optional): Test or list of tests to apply.
                Set to `'all'` to apply all the available test. Available tests:

                - `'ES'`: Epps-Singleton
                - `'KS'`: Kolmogorov-Smirnov
                - `'PSI'`: Population Stability Index
                - `'SW'`: Shapiro-Wilk
                - `'AD'`: Anderson-Darling

                Details on the available tests can be found [here](/probatus/api/stat_tests.html#available-tests).

            binning_strategies (str, optional): Binning strategies to apply for each test, either list of tests names,
                'all' or 'default'. Binning strategies that can be chosen:

                - `'SimpleBucketer'`: equally spaced bins,
                - `'AgglomerativeBucketer'`: binning by applying the Scikit-learn implementation of Agglomerative
                    Clustering,
                - `'QuantileBucketer'`: bins with equal number of elements,
                - `None`: no binning is applied. Note that not all statistical tests will be performed since some of
                    them require binning strategies.
                - `'default'`: applies a default binning for a given stats_test. For all tests apart from PSI, no
                    binning (None) is used. For PSI by default quantilebucketer is used.
                - `'all'`: each binning strategy is used for each statistical test

            bin_count (integer, None or list of integers, optional):
                bin_count value(s) to be used, note that None can only be used when no bucketing strategy is applied.
        """
        self.fitted = False

        # Initialize statistical tests to be performed
        if statistical_tests == "all":
            self.statistical_tests = list(DistributionStatistics.statistical_test_dict.keys())
        elif isinstance(statistical_tests, str):
            self.statistical_tests = [statistical_tests]
        else:
            self.statistical_tests = statistical_tests

        # Initialize binning strategies to be used
        if binning_strategies == "all":
            self.binning_strategies = list(DistributionStatistics.binning_strategy_dict.keys())
        elif isinstance(binning_strategies, str):
            self.binning_strategies = [binning_strategies]
        elif binning_strategies is None:
            self.binning_strategies = [None]
        else:
            self.binning_strategies = binning_strategies
        if not isinstance(bin_count, list):
            self.bin_count = [bin_count]
        else:
            self.bin_count = bin_count

    def __repr__(self):
        """
        String representation.
        """
        repr_ = "AutoDist object"
        if not self.fitted:
            repr_ += "\n\tAutoDist not fitted"
        if self.fitted:
            repr_ += "\n\tAutoDist fitted"
        repr_ += f"\n\tstatistical_tests: {self.statistical_tests}"
        repr_ += f"\n\tbinning_strategies: {self.binning_strategies}"
        repr_ += f"\n\tbin_count: {self.bin_count}"
        return repr_

    def compute(
        self,
        df1,
        df2,
        column_names=None,
        return_failed_tests=True,
        suppress_warnings=False,
    ):
        """
        Fit the AutoDist object to data; i.e. apply the statistical tests and binning strategies.

        Args:

            df1 (pandas.DataFrame):
                DataFrame 1 for distribution comparison with DataFrame 2.

            df2 (pandas.DataFrame):
                DataFrame 2 for distribution comparison with DataFrame 1.

            column_names (list of str, optional):
                list of columns in df1 and df2 that should be compared. If None, all column names will be compared.

            return_failed_tests (bool, optional):
                remove tests in result that did not succeed.

            suppress_warnings (bool, optional):
                whether to suppress warnings during the fit process.

        Returns:
            pandas.DataFrame: DataFrame with results of the performed statistical tests and binning strategies.

        """
        if column_names is None:
            column_names = df1.columns.to_list()
            if len(set(column_names) - set(df2.columns)):
                raise Exception("column_names was set to None but columns in provided dataframes are different")
        # Check if all columns in column_names are in df1 and df2
        elif len(set(column_names) - set(df1.columns)) or len(set(column_names) - set(df2.columns)):
            raise Exception("Not all columns in `column_names` are in the provided dataframes")

        # Calculate statistics and p-values for all combinations
        result_all = []
        for col in column_names:
            # Issue a warning if missing values are present in one of the two columns. These observations are removed
            # in the calculations.
            if np.sum(df1[col].isna()) + np.sum(df2[col].isna()):
                warnings.warn(f"Missing values in column {col} have been removed")

            # Remove the missing values.
            feature_df1 = df1[col].dropna()
            feature_df2 = df2[col].dropna()

            for stat_test, bin_strat, bins in tqdm(
                list(
                    itertools.product(
                        self.statistical_tests,
                        self.binning_strategies,
                        self.bin_count,
                    )
                )
            ):
                if self.binning_strategies == ["default"]:
                    bin_strat = DistributionStatistics.statistical_test_dict[stat_test]["default_binning"]

                dist = DistributionStatistics(
                    statistical_test=stat_test,
                    binning_strategy=bin_strat,
                    bin_count=bins,
                )
                try:
                    if suppress_warnings:
                        warnings.filterwarnings("ignore")
                    _ = dist.compute(feature_df1, feature_df2)
                    if suppress_warnings:
                        warnings.filterwarnings("default")
                    statistic = dist.statistic
                    p_value = dist.p_value
                except Exception:
                    statistic, p_value = "an error occurred", None
                    pass

                # Append result to results list
                result_ = {
                    "column": col,
                    "statistical_test": stat_test,
                    "binning_strategy": bin_strat,
                    "bin_count": bins,
                    "statistic": statistic,
                    "p_value": p_value,
                }

                result_all.append(result_)

        result_all = pd.DataFrame(result_all)

        if not return_failed_tests:
            result_all = result_all[result_all["statistic"] != "an error occurred"]
        self.fitted = True
        self._result = result_all[
            [
                "column",
                "statistical_test",
                "binning_strategy",
                "bin_count",
                "statistic",
                "p_value",
            ]
        ]
        self._result["bin_count"] = self._result["bin_count"].astype(int)
        self._result.loc[self._result["binning_strategy"].isnull(), "bin_count"] = 0
        self._result.loc[self._result["binning_strategy"].isnull(), "binning_strategy"] = "no_bucketing"

        # Remove duplicates that appear if multiple bin numbers are passed, and binning strategy None

        self._result = self._result.drop_duplicates(
            subset=["column", "statistical_test", "binning_strategy", "bin_count"],
            keep="first",
        )

        # create pivot table as final output
        self.result = pd.pivot_table(
            self._result,
            values=["statistic", "p_value"],
            index="column",
            columns=["statistical_test", "binning_strategy", "bin_count"],
            aggfunc="sum",
        )

        # flatten multi-index
        self.result.columns = ["_".join([str(x) for x in line]) for line in self.result.columns.values]
        self.result.reset_index(inplace=True)
        return self.result



import os
from unittest.mock import patch

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest
from sklearn.tree import DecisionTreeClassifier

from probatus.metric_volatility import (
    BaseVolatilityEstimator,
    BootstrappedVolatility,
    SplitSeedVolatility,
    TrainTestVolatility,
    check_sampling_input,
    get_metric,
    sample_data,
)
from probatus.stat_tests.distribution_statistics import DistributionStatistics
from probatus.utils import NotFittedError, Scorer

# Turn off interactive mode in plots
plt.ioff()
matplotlib.use("Agg")


@pytest.fixture(scope="function")
def X_array():
    """
    Fixture.
    """
    return np.array([[2, 1], [3, 2], [4, 3], [1, 2], [1, 1]])


@pytest.fixture(scope="function")
def y_list():
    """
    Fixture.
    """
    return [1, 0, 0, 1, 1]


@pytest.fixture(scope="function")
def y_array(y_list):
    """
    Fixture.
    """
    return np.array(y_list)


@pytest.fixture(scope="function")
def X_df(X_array):
    """
    Fixture.
    """
    return pd.DataFrame(X_array, columns=["c1", "c2"])


@pytest.fixture(scope="function")
def y_series(y_list):
    """
    Fixture.
    """
    return pd.Series(y_list)


@pytest.fixture(scope="function")
def iteration_results():
    """
    Fixture.
    """
    iterations_cols = ["metric_name", "train_score", "test_score", "delta_score"]
    return pd.DataFrame(
        [
            ["roc_auc", 0.8, 0.7, 0.1],
            ["roc_auc", 0.7, 0.6, 0.1],
            ["roc_auc", 0.9, 0.8, 0.1],
            ["accuracy", 1, 0.9, 0.1],
            ["accuracy", 0.8, 0.7, 0.1],
            ["accuracy", 0.9, 0.8, 0.1],
        ],
        columns=iterations_cols,
    )


@pytest.fixture(scope="function")
def report():
    """
    Fixture.
    """
    report_cols = ["train_mean", "train_std", "test_mean", "test_std", "delta_mean", "delta_std"]
    report_index = ["roc_auc", "accuracy"]
    return pd.DataFrame(
        [[0.8, 0.08164, 0.7, 0.08164, 0.1, 0], [0.9, 0.08164, 0.8, 0.08164, 0.1, 0]],
        columns=report_cols,
        index=report_index,
    ).astype(float)


@pytest.fixture(scope="function")
def iterations_train():
    """
    Fixture.
    """
    return pd.Series([0.8, 0.7, 0.9], name="train_score")


@pytest.fixture(scope="function")
def iterations_test():
    """
    Fixture.
    """
    return pd.Series([0.7, 0.6, 0.8], name="test_score")


@pytest.fixture(scope="function")
def iterations_delta():
    """
    Fixture.
    """
    return pd.Series([0.1, 0.1, 0.1], name="delta_score")


def test_inits(mock_model):
    """
    Test.
    """
    vol1 = SplitSeedVolatility(
        mock_model,
        scoring=["accuracy", "roc_auc"],
        test_prc=0.3,
        n_jobs=2,
        stats_tests_to_apply=["ES", "KS"],
        random_state=1,
        iterations=20,
    )

    assert id(vol1.clf) == id(mock_model)
    assert vol1.test_prc == 0.3
    assert vol1.n_jobs == 2
    assert vol1.stats_tests_to_apply == ["ES", "KS"]
    assert vol1.random_state == 1
    assert vol1.iterations == 20
    assert len(vol1.stats_tests_objects) == 2
    assert len(vol1.scorers) == 2
    assert vol1.sample_train_test_split_seed is True

    vol2 = BootstrappedVolatility(mock_model, scoring="roc_auc", stats_tests_to_apply="KS", test_sampling_fraction=0.8)

    assert id(vol2.clf) == id(mock_model)
    assert vol2.stats_tests_to_apply == ["KS"]
    assert len(vol2.stats_tests_objects) == 1
    assert len(vol2.scorers) == 1
    assert vol2.sample_train_test_split_seed is False
    assert vol2.test_sampling_fraction == 0.8
    assert vol2.fitted is False
    assert vol2.iterations_results is None
    assert vol2.report is None


def test_base_fit(mock_model, X_df, y_series):
    """
    Test.
    """
    vol = BaseVolatilityEstimator(mock_model, random_state=1)

    with patch("numpy.random.seed") as mock_seed:
        vol.fit(X_df, y_series)
        mock_seed.assert_called_with(1)

    assert vol.iterations_results is None
    assert vol.report is None
    assert vol.fitted is True


def test_compute(report, mock_model):
    """
    Test.
    """
    vol = BaseVolatilityEstimator(mock_model)

    with pytest.raises(NotFittedError):
        vol.compute()

    vol.fit()
    with pytest.raises(ValueError):
        vol.compute()

    vol.report = report

    pd.testing.assert_frame_equal(vol.compute(), report)
    pd.testing.assert_frame_equal(vol.compute(metrics=["roc_auc"]), report.loc[["roc_auc"]])
    pd.testing.assert_frame_equal(vol.compute(metrics="roc_auc"), report.loc[["roc_auc"]])


def test_plot(report, mock_model, iterations_train, iterations_test, iterations_delta):
    """
    Test.
    """
    with patch.object(BaseVolatilityEstimator, "compute", return_value=report.loc[["roc_auc"]]) as mock_compute:
        with patch.object(
            BaseVolatilityEstimator,
            "_get_samples_to_plot",
            return_value=(iterations_train, iterations_test, iterations_delta),
        ) as mock_get_samples:
            vol = BaseVolatilityEstimator(mock_model)
            vol.fitted = True

            vol.plot(metrics="roc_auc")
            mock_compute.assert_called_with(metrics="roc_auc")
            mock_get_samples.assert_called_with(metric_name="roc_auc")


def test_get_samples_to_plot(mock_model, iteration_results, iterations_train, iterations_test, iterations_delta):
    """
    Test.
    """
    vol = BaseVolatilityEstimator(mock_model)
    vol.fitted = True
    vol.iterations_results = iteration_results

    train, test, delta = vol._get_samples_to_plot(metric_name="roc_auc")
    pd.testing.assert_series_equal(train, iterations_train)
    pd.testing.assert_series_equal(test, iterations_test)
    pd.testing.assert_series_equal(delta, iterations_delta)


def test_create_report(mock_model, iteration_results, report):
    """
    Test.
    """
    vol = BaseVolatilityEstimator(mock_model)
    vol.fitted = True
    vol.iterations_results = iteration_results

    vol._create_report()
    pd.testing.assert_frame_equal(vol.report, report, atol=1e-3)


def test_compute_mean_std_from_runs(mock_model, iteration_results):
    """
    Test.
    """
    vol = BaseVolatilityEstimator(mock_model)
    results = vol._compute_mean_std_from_runs(iteration_results[iteration_results["metric_name"] == "roc_auc"])
    expected_results = [0.8, 0.08164, 0.7, 0.08164, 0.1, 0]
    for idx, item in enumerate(results):
        assert pytest.approx(item, 0.01) == expected_results[idx]


def test_compute_stats_tests_values(mock_model, iteration_results):
    """
    Test.
    """
    vol = BaseVolatilityEstimator(mock_model, stats_tests_to_apply=["KS"])

    with patch.object(DistributionStatistics, "compute", return_value=(0.1, 0.05)):
        stats = vol._compute_stats_tests_values(iteration_results)

    assert stats[0] == 0.1
    assert stats[1] == 0.05


def test_fit_compute(mock_model, report, X_df, y_series):
    """
    Test.
    """
    vol = BaseVolatilityEstimator(mock_model)

    with patch.object(BaseVolatilityEstimator, "fit") as mock_fit:
        with patch.object(BaseVolatilityEstimator, "compute", return_value=report) as mock_compute:
            result = vol.fit_compute(X_df, y_series)

            mock_fit.assert_called_with(X_df, y_series)
            mock_compute.assert_called_with()

    pd.testing.assert_frame_equal(result, report)


def test_fit_train_test_sample_seed(mock_model, X_df, y_series, iteration_results):
    """
    Test.
    """
    vol = TrainTestVolatility(mock_model, scoring="roc_auc", iterations=3, sample_train_test_split_seed=True)

    with patch.object(BaseVolatilityEstimator, "fit") as mock_base_fit:
        with patch.object(TrainTestVolatility, "_create_report") as mock_create_report:
            with patch(
                "probatus.metric_volatility.volatility.get_metric",
                side_effect=[iteration_results.iloc[[0]], iteration_results.iloc[[1]], iteration_results.iloc[[2]]],
            ):
                vol.fit(X_df, y_series)

                mock_base_fit.assert_called_once()
                mock_create_report.assert_called_once()

    pd.testing.assert_frame_equal(vol.iterations_results, iteration_results.iloc[[0, 1, 2]])


def test_get_metric(mock_model, X_df, y_series):
    """
    Test.
    """
    split_seed = 1
    test_prc = 0.6
    with patch(
        "probatus.metric_volatility.metric.train_test_split",
        return_value=(X_df.iloc[[0, 1, 2]], X_df.iloc[[3, 4]], y_series.iloc[[0, 1, 2]], y_series.iloc[[3, 4]]),
    ) as mock_split:
        with patch(
            "probatus.metric_volatility.metric.sample_data",
            side_effect=[(X_df.iloc[[0, 1, 1]], y_series.iloc[[0, 1, 1]]), (X_df.iloc[[3, 3]], y_series.iloc[[3, 3]])],
        ) as mock_sample:
            with patch.object(Scorer, "score", side_effect=[0.8, 0.7]):
                output = get_metric(
                    X_df,
                    y_series,
                    mock_model,
                    test_size=test_prc,
                    split_seed=split_seed,
                    scorers=[Scorer("roc_auc")],
                    train_sampling_type="bootstrap",
                    test_sampling_type="bootstrap",
                    train_sampling_fraction=1,
                    test_sampling_fraction=1,
                )
                mock_split.assert_called_once()
                mock_sample.assert_called()
                mock_model.fit.assert_called()

    expected_output = pd.DataFrame(
        [["roc_auc", 0.8, 0.7, 0.1]], columns=["metric_name", "train_score", "test_score", "delta_score"]
    )
    pd.testing.assert_frame_equal(expected_output, output)


def test_sample_data_no_sampling(X_df, y_series):
    """
    Test.
    """
    with patch("probatus.metric_volatility.utils.check_sampling_input") as mock_sampling_input:
        X_out, y_out = sample_data(X_df, y_series, sampling_type=None, sampling_fraction=1)
        mock_sampling_input.assert_called_once()
    pd.testing.assert_frame_equal(X_out, X_df)
    pd.testing.assert_series_equal(y_out, y_series)


def test_sample_data_bootstrap(X_df, y_series):
    """
    Test.
    """
    with patch("probatus.metric_volatility.utils.check_sampling_input") as mock_sampling_input:
        X_out, y_out = sample_data(X_df, y_series, sampling_type="bootstrap", sampling_fraction=0.8)
        mock_sampling_input.assert_called_once()
    assert X_out.shape == (4, 2)
    assert y_out.shape == (4,)


def test_sample_data_sample(X_df, y_series):
    """
    Test.
    """
    with patch("probatus.metric_volatility.utils.check_sampling_input") as mock_sampling_input:
        X_out, y_out = sample_data(X_df, y_series, sampling_type="subsample", sampling_fraction=1)
        mock_sampling_input.assert_called_once()
    pd.testing.assert_frame_equal(X_out, X_df)
    pd.testing.assert_series_equal(y_out, y_series)


def test_check_sampling_input(X_array, y_array):
    """
    Test.
    """
    with pytest.raises(ValueError):
        check_sampling_input("bootstrap", 0, "dataset")
    with pytest.raises(ValueError):
        check_sampling_input("subsample", 0, "dataset")
    with pytest.raises(ValueError):
        check_sampling_input("subsample", 1, "dataset")
    with pytest.raises(ValueError):
        check_sampling_input("subsample", 10, "dataset")
    with pytest.raises(ValueError):
        check_sampling_input("wrong_name", 0.5, "dataset")


def test_fit_compute_full_process(X_df, y_series):
    """
    Test.
    """
    clf = DecisionTreeClassifier()
    vol = TrainTestVolatility(
        clf, scoring=["roc_auc", "recall"], iterations=3, sample_train_test_split_seed=False, random_state=42
    )

    report = vol.fit_compute(X_df, y_series)
    assert report.shape == (2, 6)

    # Check if plot runs
    vol.plot(show=False)


@pytest.mark.skipif(os.environ.get("SKIP_LIGHTGBM") == "true", reason="LightGBM tests disabled")
def test_fit_compute_complex(complex_data, complex_lightgbm):
    """
    Test.
    """
    X, y = complex_data
    vol = TrainTestVolatility(
        complex_lightgbm,
        scoring="roc_auc",
        iterations=3,
        sample_train_test_split_seed=True,
        verbose=150,
        random_state=42,
    )

    report = vol.fit_compute(X, y)
    assert report.shape == (1, 6)

    # Check if plot runs
    vol.plot(show=False)



# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from probatus.stat_tests.utils import verbose_p_vals
from probatus.utils import NotInstalledError

from ..utils import assure_numpy_array

try:
    from scipy import stats
except ModuleNotFoundError:
    stats = NotInstalledError("scipy", "extras")


@verbose_p_vals
def ad(d1, d2, verbose=False):
    """
    Calculates the Anderson-Darling test statistic on 2 distributions.

    Can be used on continuous or discrete distributions.

    Any binning/bucketing of the distributions/samples should be done before passing them to this function.

    Advantages:

    - Unlike the KS, the AD (like the ES) can be used on both continuous & discrete distributions.
    - Works well even when the sample has fewer than 25 observations.
    - More powerful than KS, especially for differences in the tails of distributions.

    References:

    - [Wikipedia article about the Anderson-Darling test](https://en.wikipedia.org/wiki/Anderson%E2%80%93Darling_test)
    - [SciPy documentation](https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.anderson_ksamp.html)

    Args:
        d1 (np.array or pandas.Series): First sample.

        d2 (np.array or pandas.Series): Second sample.

        verbose (bool): If True, useful interpretation info is printed to stdout.

    Returns:
        float: Anderson-Darling test statistic.
        float: p-value of rejecting the null hypothesis (that the two distributions are identical).
    """
    d1 = assure_numpy_array(d1)
    d2 = assure_numpy_array(d2)

    ad, critical_values, pvalue = stats.anderson_ksamp([d1, d2])

    return ad, pvalue



import numbers

import numpy as np
import pandas as pd
import pytest
from sklearn.datasets import make_classification
from sklearn.model_selection import train_test_split

from probatus.stat_tests import AutoDist, DistributionStatistics, ks, psi


def test_distribution_statistics_base():
    """
    Test.
    """
    with pytest.raises(NotImplementedError):
        assert DistributionStatistics("doesnotexist", "SimpleBucketer", bin_count=10)
    with pytest.raises(NotImplementedError):
        assert DistributionStatistics("psi", "doesnotexist", bin_count=10)
    myTest = DistributionStatistics("psi", "SimpleBucketer", bin_count=10)
    assert repr(myTest).startswith("DistributionStatistics")


def test_distribution_statistics_psi():
    """
    Test.
    """
    d1 = np.histogram(np.random.normal(size=1000), 10)[0]
    d2 = np.histogram(np.random.weibull(1, size=1000) - 1, 10)[0]
    myTest = DistributionStatistics("psi", "SimpleBucketer", bin_count=10)
    assert not myTest.fitted
    psi_test, p_value_test = myTest.compute(d1, d2)
    assert myTest.fitted
    assert isinstance(psi_test, numbers.Number)


def test_distribution_statistics_tuple_output():
    """
    Test.
    """
    d1 = np.histogram(np.random.normal(size=1000), 10)[0]
    d2 = np.histogram(np.random.weibull(1, size=1000) - 1, 10)[0]
    myTest = DistributionStatistics("ks", "SimpleBucketer", bin_count=10)
    assert not myTest.fitted
    res = myTest.compute(d1, d2)
    assert myTest.fitted
    assert isinstance(res, tuple)


def test_distribution_statistics_ks_no_binning():
    """
    Test.
    """
    d1 = np.histogram(np.random.normal(size=1000), 10)[0]
    d2 = np.histogram(np.random.weibull(1, size=1000) - 1, 10)[0]
    myTest = DistributionStatistics("ks", binning_strategy=None)
    assert not myTest.fitted
    res = myTest.compute(d1, d2)
    assert myTest.fitted
    assert isinstance(res, tuple)


def test_distribution_statistics_attributes_psi():
    """
    Test.
    """
    a = np.random.normal(size=1000)
    b = np.random.normal(size=1000)
    d1 = np.histogram(a, 10)[0]
    d2 = np.histogram(b, 10)[0]
    myTest = DistributionStatistics("psi", binning_strategy=None)
    _ = myTest.compute(d1, d2, verbose=False)
    psi_value_test, p_value_test = psi(d1, d2, verbose=False)
    assert myTest.statistic == psi_value_test


def test_distribution_statistics_attributes_ks():
    """
    Test.
    """
    d1 = np.histogram(np.random.normal(size=1000), 10)[0]
    d2 = np.histogram(np.random.normal(size=1000), 10)[0]
    myTest = DistributionStatistics("ks", binning_strategy=None)
    _ = myTest.compute(d1, d2, verbose=False)
    ks_value, p_value = ks(d1, d2)
    assert myTest.statistic == ks_value


def test_distribution_statistics_autodist_base():
    """
    Test.
    """
    nr_features = 2
    size = 1000
    np.random.seed(0)
    df1 = pd.DataFrame(np.random.normal(size=(size, nr_features)), columns=[f"feat_{x}" for x in range(nr_features)])
    df2 = pd.DataFrame(np.random.normal(size=(size, nr_features)), columns=[f"feat_{x}" for x in range(nr_features)])
    features = df1.columns
    myAutoDist = AutoDist(statistical_tests="all", binning_strategies="all", bin_count=[10, 20])
    assert repr(myAutoDist).startswith("AutoDist")
    assert not myAutoDist.fitted
    res = myAutoDist.compute(df1, df2, column_names=features)
    assert myAutoDist.fitted
    pd.testing.assert_frame_equal(res, myAutoDist.result)
    assert isinstance(res, pd.DataFrame)
    assert res["column"].values.tolist() == features.to_list()

    dist = DistributionStatistics(statistical_test="ks", binning_strategy="simplebucketer", bin_count=10)
    dist.compute(df1["feat_0"], df2["feat_0"])
    assert dist.p_value == res.loc[res["column"] == "feat_0", "p_value_KS_simplebucketer_10"][0]
    assert dist.statistic == res.loc[res["column"] == "feat_0", "statistic_KS_simplebucketer_10"][0]

    dist = DistributionStatistics(statistical_test="ks", binning_strategy=None, bin_count=10)
    dist.compute(df1["feat_0"], df2["feat_0"])
    assert dist.p_value == res.loc[res["column"] == "feat_0", "p_value_KS_no_bucketing_0"][0]
    assert dist.statistic == res.loc[res["column"] == "feat_0", "statistic_KS_no_bucketing_0"][0]


def test_distribution_statistics_autodist_column_names_error():
    """
    Test.
    """
    df1 = pd.DataFrame({"feat_0": [1, 2, 3, 4, 5], "feat_1": [5, 6, 7, 8, 9]})
    df2 = df1
    features = df1.columns.values.tolist() + ["missing_feature"]
    myAutoDist = AutoDist()
    with pytest.raises(Exception):
        assert myAutoDist.compute(df1, df2, column_names=features)

    df1 = pd.DataFrame({"feat_0": [1, 2, 3, 4, 5], "feat_1": [5, 6, 7, 8, 9]})
    df2 = df1.copy()
    df1["feat_2"] = 0
    features = df2.columns.values.tolist() + ["missing_feature"]
    myAutoDist = AutoDist()
    with pytest.raises(Exception):
        assert myAutoDist.compute(df1, df2, column_names=features)


@pytest.mark.skip(reason="Currently fails on ubuntu, to be investigated further.")
def test_distribution_statistics_autodist_return_failed_tests():
    """
    Test.
    """
    df1 = pd.DataFrame({"feat_0": [1, 2, 3, 4, 5], "feat_1": [5, 6, 7, 8, 9]})
    df2 = df1
    features = df1.columns.values.tolist()
    myAutoDist = AutoDist(binning_strategies="all")
    res = myAutoDist.compute(df1, df2, column_names=features, return_failed_tests=True)
    assert res.isin(["an error occurred"]).any().any()
    res = myAutoDist.compute(df1, df2, column_names=features, return_failed_tests=False)
    assert not res.isin(["an error occurred"]).any().any()


def test_distribution_statistics_autodist_default():
    """
    Test.
    """
    df1 = pd.DataFrame({"feat_0": [1, 2, 3, 4, 5], "feat_1": [5, 6, 7, 8, 9]})
    df2 = df1
    features = df1.columns.values.tolist()
    myAutoDist = AutoDist(binning_strategies="default", bin_count=10)
    res = myAutoDist.compute(df1, df2, column_names=features)
    for stat_test, stat_info in DistributionStatistics.statistical_test_dict.items():
        if stat_info["default_binning"]:
            assert f"p_value_{stat_test}_{stat_info['default_binning']}_10" in res.columns
        else:
            assert f"p_value_{stat_test}_no_bucketing_0" in res.columns

    assert "p_value_agglomerativebucketer_10" not in res.columns
    assert res.shape == (len(df1.columns), 1 + 2 * len(DistributionStatistics.statistical_test_dict))


def test_distribution_statistics_autodist_init():
    """
    Test.
    """
    myAutoDist = AutoDist(statistical_tests="all", binning_strategies="all")
    assert isinstance(myAutoDist.statistical_tests, list)
    myAutoDist = AutoDist(statistical_tests="ks", binning_strategies="all")
    assert myAutoDist.statistical_tests == ["ks"]
    myAutoDist = AutoDist(statistical_tests=["ks", "psi"], binning_strategies="all")
    assert myAutoDist.statistical_tests == ["ks", "psi"]

    myAutoDist = AutoDist(statistical_tests="all", binning_strategies="all")
    assert isinstance(myAutoDist.binning_strategies, list)
    myAutoDist = AutoDist(statistical_tests="all", binning_strategies="quantilebucketer")
    assert myAutoDist.binning_strategies == ["quantilebucketer"]
    myAutoDist = AutoDist(statistical_tests="all", binning_strategies=["quantilebucketer", "simplebucketer"])
    assert myAutoDist.binning_strategies == ["quantilebucketer", "simplebucketer"]


def test_missing_values_in_autodist():
    """Test missing values have no impact in AutoDist functionality."""
    # Create dummy dataframe
    X, y = make_classification(50, 5, random_state=0)
    X = pd.DataFrame(X)
    # Split train and test
    X_train, X_test, _, _ = train_test_split(X, y, test_size=0.2, random_state=1)
    # Define an add-on with only missing values
    X_na = pd.DataFrame(np.tile(np.nan, (X.shape[1], X.shape[1])))

    # Compute the statistics with the missing values
    with_missings = AutoDist(
        statistical_tests=["PSI", "KS"], binning_strategies="SimpleBucketer", bin_count=10
    ).compute(pd.concat([X_train, X_na]), pd.concat([X_test, X_na]))

    # Compute the statistics withpout the missing values
    no_missing = AutoDist(statistical_tests=["PSI", "KS"], binning_strategies="SimpleBucketer", bin_count=10).compute(
        X_train, X_test
    )

    # Test the two set of results are identical
    pd.testing.assert_frame_equal(with_missings, no_missing)


def test_warnings_are_issued_for_missing():
    """Test if warnings are issued when missing values are present in the input of autodist."""
    # Generate an input dataframe without missing values
    X = pd.DataFrame({"A": [number for number in range(0, 50)]})
    X = X.assign(B=X["A"], C=X["A"], D=X["A"], E=X["A"])

    # Add some missing values to the dataframe.
    X_na = X.copy()
    X_na.iloc[X.sample(5, random_state=1).index, 1:3] = np.nan

    # Test missing value removal on the first data input.
    with pytest.warns(None) as record_first:
        _ = AutoDist(statistical_tests=["PSI"], binning_strategies="SimpleBucketer", bin_count=10).compute(X_na, X)
    assert len(record_first) == 2

    # Test missing values removal on the second data input
    with pytest.warns(None) as record_second:
        _ = AutoDist(statistical_tests=["PSI"], binning_strategies="SimpleBucketer", bin_count=10).compute(X, X_na)
    assert len(record_second) == 2

    # Test the missing values removal on the first and second data input
    with pytest.warns(None) as record_both:
        _ = AutoDist(statistical_tests=["PSI"], binning_strategies="SimpleBucketer", bin_count=10).compute(X_na, X_na)
    assert len(record_both) == 2

    # Test case where there are no missing values
    with pytest.warns(None) as record_both:
        _ = AutoDist(statistical_tests=["PSI"], binning_strategies="SimpleBucketer", bin_count=10).compute(X, X)
    assert len(record_both) == 0



# Copyright (c) 2021 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.compose import ColumnTransformer
from sklearn.impute import SimpleImputer
from sklearn.model_selection import cross_validate
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder

from probatus.utils import BaseFitComputePlotClass, get_single_scorer, preprocess_data, preprocess_labels


class ImputationSelector(BaseFitComputePlotClass):
    """
    Comparison of various imputation strategies that can be used for imputing missing values.

    The aim of this class is to present the model performance based on imputation
    strategies and a chosen model.
    For models like XGBoost & LighGBM which have capabilities to handle missing values by default
    the model performance with no imputation will be shown as well.
    The missing values categorical features are imputed with the value `missing` and an missing indicator is
    added.

    Example:
    ```python

    #Import the class
    import pandas as pd
    import numpy as np
    import matplotlib.pyplot as plt
    from probatus.missing_values.imputation import ImputationSelector
    from probatus.utils.missing_helpers import generate_MCAR
    from sklearn.linear_model import LogisticRegression
    from sklearn.experimental import enable_iterative_imputer
    from sklearn.impute import KNNImputer,SimpleImputer,IterativeImputer
    from sklearn.datasets import make_classification

    # Create data with missing values.
    n_features = 10
    X,y = make_classification(n_samples=1000,n_features=n_features,random_state=123,class_sep=0.3)
    X = pd.DataFrame(X, columns=["f_"+str(i) for i in range(0,n_features)])
    X_missing = generate_MCAR(X,missing=0.2)

    # Create the strategies.
    strategies = {
       'Simple Median Imputer' : SimpleImputer(strategy='median',add_indicator=True),
       'Simple Mean Imputer' : SimpleImputer(strategy='mean',add_indicator=True),
       'Iterative Imputer'  : IterativeImputer(add_indicator=True,n_nearest_features=5,
       sample_posterior=True),
       'KNN' : KNNImputer(n_neighbors=3)}
    #Create a classifier.
    clf = LogisticRegression()
    #Create the comparison of the imputation strategies.
    cmp = ImputationSelector(
        clf=clf,
        strategies=strategies,
        cv=5,
        model_na_support=False)

    cmp.fit_compute(X_missing,y)
    #Plot the results.
    performance_plot=cmp.plot()

    ```

    <img src="../img/imputation_comparison.png" width="500" />

    """

    def __init__(
        self,
        clf,
        strategies,
        scoring="roc_auc",
        cv=5,
        model_na_support=False,
        n_jobs=-1,
        verbose=0,
        random_state=None,
    ):
        """
        Initialise the class.

        Args:
            clf (binary classifier,sklearn.Pipeline):
                A binary classification model, that will used to evaluate various imputation strategies.

            strategies (dictionary of sklearn.impute objects or any other scikit learn compatible imputer.):
                Dictionary containing the sklearn.impute objects.
                e.g.
                strategies = {'KNN' : KNNImputer(n_neighbors=3),
                'Simple Median Imputer' : SimpleImputer(strategy='median',add_indicator=True),
                'Iterative Imputer'  : IterativeImputer(add_indicator=True,n_nearest_features=5,
                sample_posterior=True)}
                This allows you to have fine grained control over the imputation method.

            scoring (string, list of strings, probatus.utils.Scorer or list of probatus.utils.Scorers, optional):
                Metrics for which the score is calculated. It can be either a name or list of names metric names and
                needs to be aligned with predefined
                [classification scorers names in sklearn](https://scikit-learn.org/stable/modules/model_evaluation.html).
                Another option is using probatus.utils.Scorer to define a custom metric.

            model_na_support (boolean): default False
                If the classifier supports missing values by default e.g. LightGBM,XGBoost etc.
                If True an default comparison `No Imputation`  result will be added indicating the model performance
                    without any explicit imputation.
                If False only the provided strategies will be used.

            n_jobs (int, optional):
                Number of cores to run in parallel while fitting across folds. None means 1 unless in a
                `joblib.parallel_backend` context. -1 means using all processors.

            verbose (int, optional):
                Controls verbosity of the output:

                - 0 - nether prints nor warnings are shown
                - 1 - 50 - only most important warnings regarding data properties are shown (excluding SHAP warnings)
                - 51 - 100 - shows most important warnings, prints of the feature removal process
                - above 100 - presents all prints and all warnings (including SHAP warnings).

            random_state (int, optional):
                Random state set at each round of feature elimination. If it is None, the results will not be
                reproducible and in random search at each iteration a different hyperparameters might be tested. For
                reproducible results set it to integer.
        """  # noqa
        self.clf = clf
        self.model_na_support = model_na_support
        self.cv = cv
        self.scorer = get_single_scorer(scoring)
        self.strategies = strategies
        self.verbose = verbose
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.fitted = False
        self.report_df = pd.DataFrame([])

    def __repr__(self):
        """
        String representation.
        """
        return f"Imputation comparison for {self.clf.__class__.__name__}"

    def fit(self, X, y, column_names=None):
        """
        Calculates the cross validated results for various imputation strategies.

        Args:
            X (pd.DataFrame):
                input variables.

            y (pd.Series):
                target variable.

            column_names (None, or list of str, optional):
                List of feature names for the dataset.
                If None, then column names from the X dataframe are used.
        """
        if self.random_state is not None:
            np.random.seed(self.random_state)

        # Place holder for results.
        results = []

        self.X, self.column_names = preprocess_data(X, column_names=column_names, verbose=self.verbose)
        self.y = preprocess_labels(y, index=self.X.index, verbose=self.verbose)

        # Identify categorical features.
        categorical_columns = X.select_dtypes(include=["category", "object"]).columns
        # Identify the numeric columns.Numeric columns are all columns expect the categorical columns
        numeric_columns = X.select_dtypes("number").columns

        for strategy in self.strategies:
            numeric_transformer = Pipeline(steps=[("imputer", self.strategies[strategy])])

            categorical_transformer = Pipeline(
                steps=[
                    (
                        "imp_cat",
                        SimpleImputer(
                            strategy="constant",
                            fill_value="missing",
                            add_indicator=True,
                        ),
                    ),
                    ("ohe_cat", OneHotEncoder(handle_unknown="ignore")),
                ]
            )

            preprocessor = ColumnTransformer(
                transformers=[
                    ("num", numeric_transformer, numeric_columns),
                    ("cat", categorical_transformer, categorical_columns),
                ],
                remainder="passthrough",
            )

            model_pipeline = Pipeline(steps=[("preprocessor", preprocessor), ("classifier", self.clf)])

            temp_results = self._calculate_results(X, y, clf=model_pipeline, strategy=strategy)

            results.append(temp_results)

        # If model supports missing values by default, then calculate the scores
        # on raw data without any imputation.
        if self.model_na_support:
            categorical_transformer = Pipeline(
                steps=[
                    ("ohe_cat", OneHotEncoder(handle_unknown="ignore")),
                ]
            )

            preprocessor = ColumnTransformer(
                transformers=[("cat", categorical_transformer, categorical_columns)],
                remainder="passthrough",
            )

            model_pipeline = Pipeline(steps=[("preprocessor", preprocessor), ("classifier", self.clf)])

            temp_results = self._calculate_results(X, y, clf=model_pipeline, strategy="No Imputation")
            results.append(temp_results)

        self.report_df = pd.DataFrame(results)
        # Set the index of the dataframe to the imputation methods.
        self.report_df = self.report_df.set_index(self.report_df.strategy)
        self.report_df.drop(columns=["strategy"], inplace=True)
        self.report_df.sort_values(by="mean_test_score", inplace=True)
        self.fitted = True
        return self

    def _calculate_results(self, X, y, clf, strategy):
        """
        Method to calculate the results for a particular imputation strategy.

        Args:
            X (pd.DataFrame):
                input variables.

            y (pd.Series):
                target variable.

            clf (binary classifier,sklearn.Pipeline):
                A binary classification model, that will used to evaluate various imputation strategies.

            strategy(string):
                Name of the strategy used for imputation.

        Returns:

            temp_df(dict) :  Dictionary containing the results of the evaluation.
        """

        imputation_cv_results = cross_validate(
            clf,
            X,
            y,
            scoring=self.scorer.scorer,
            cv=self.cv,
            n_jobs=self.n_jobs,
            return_train_score=True,
        )
        # Calculate the mean of the results.
        imp_agg_results = {k: np.mean(v) for k, v in imputation_cv_results.items()}
        imp_agg_results = {"mean_" + str(key): val for key, val in imp_agg_results.items()}
        imp_agg_results["test_score_std"] = np.std(imputation_cv_results["test_score"])
        imp_agg_results["train_score_std"] = np.std(imputation_cv_results["train_score"])
        # Round off all calculations to 3 decimal places
        imp_agg_results = {k: np.round(v, 3) for k, v in imp_agg_results.items()}
        imp_agg_results["strategy"] = strategy

        return imp_agg_results

    def compute(self):
        """
        Checks if fit() method has been run.

        and computes the DataFrame with results of imputation for each
        strategy.

        Returns:
            (pd.DataFrame):
                 DataFrame with results of imputation for each strategy.
        """
        self._check_if_fitted()
        return self.report_df

    def fit_compute(self, X, y, column_names=None):
        """
        Calculates the cross validated results for various imputation strategies.

        Args:
            X (pd.DataFrame):
                input variables.

            y (pd.Series):
                target variable.

            column_names (None, or list of str, optional):
                List of feature names for the dataset.
                If None, then column names from the X dataframe are used.

        Returns:
            (pd.DataFrame):
                DataFrame with results of imputation for each strategy.

        """
        self.fit(X, y, column_names=column_names)
        return self.compute()

    def plot(self, show=True, **figure_kwargs):
        """
        Generates plot of the performance of various imputation strategies.

        Args:
            show (bool, optional):
                If True, the plots are showed to the user, otherwise they are not shown. Not showing plot can be useful,
                when you want to edit the returned axis, before showing it.

            **figure_kwargs:
                Keyword arguments that are passed to the plt.figure, at its initialization.

        Returns:
            (plt.axis):
                Axis containing the performance plot.
        """
        fig, ax = plt.subplots(**figure_kwargs)

        report_df = self.compute()
        imp_methods = list(report_df.index)
        test_performance = list(report_df["mean_test_score"])
        test_std_error = list(report_df["test_score_std"])
        train_performance = list(report_df["mean_train_score"])
        train_std_error = list(report_df["train_score_std"])

        y = np.arange(len(imp_methods))  # the label locations
        width = 0.35  # the width of the bars

        def _autolabel(rects):
            """
            Label the bars of the plot.
            """
            for rect in rects:
                width = rect.get_width()
                ax.annotate(
                    f"{width}",
                    xy=((width + 0.05 * width), rect.get_y() + rect.get_height() / 2),
                    xytext=(4, 0),  # 4 points horizontal offset
                    textcoords="offset points",
                    ha="center",
                    va="bottom",
                    fontsize="small",
                )

        train_rect = ax.barh(
            y - width / 2,
            train_performance,
            width,
            xerr=train_std_error,
            align="center",
            label="CV-Train",
        )
        test_rect = ax.barh(
            y + width / 2,
            test_performance,
            width,
            xerr=test_std_error,
            align="center",
            label="CV-Test",
        )
        _autolabel(train_rect)
        _autolabel(test_rect)

        ax.set_xlabel(f'{self.scorer.metric_name.replace("_"," ").upper()} Score')
        ax.set_title("Imputation Techniques Comparison")
        ax.set_yticks(y)
        ax.set_yticklabels(imp_methods, rotation=45)
        plt.margins(0.2)
        plt.legend(loc="best", ncol=2)
        fig.tight_layout()

        if show:
            plt.show()
        else:
            plt.close()
        return ax



# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import functools


def verbose_p_vals(func):
    """
    Decorator to enable verbose printing of p-values.
    """

    @functools.wraps(func)
    def wrapper_verbose_p_vals(*args, **kwargs):
        test_name = func.__name__.upper()

        stat, pvalue = func(*args, **kwargs)

        if "verbose" in kwargs and kwargs["verbose"] is True:
            print(f"\n{test_name}: pvalue =", pvalue)
            if pvalue < 0.01:
                print(
                    "\n{}: Null hypothesis rejected with 99% confidence. Distributions very different.".format(
                        test_name
                    )
                )
            elif pvalue < 0.05:
                print(f"\n{test_name}: Null hypothesis rejected with 95% confidence. Distributions different.")
            else:
                print(
                    "\n{}: Null hypothesis cannot be rejected. Distributions not statistically different.".format(
                        test_name
                    )
                )

        return stat, pvalue

    return wrapper_verbose_p_vals



# Code to test the imputation strategies.
import os

import numpy as np
import pandas as pd
import pytest
from sklearn.ensemble import RandomForestClassifier
from sklearn.experimental import enable_iterative_imputer  # noqa
from sklearn.impute import IterativeImputer, KNNImputer, SimpleImputer
from sklearn.linear_model import LogisticRegression

from probatus.missing_values.imputation import ImputationSelector


@pytest.fixture(scope="function")
def X():
    """
    Fixture.
    """
    return pd.DataFrame(
        {
            "col_1": [1, np.nan, 1, 1, np.nan, 1, 1, 0, 1, 1],
            "col_2": [0, 0, 0, np.nan, 0, 0, 0, 1, 0, 0],
            "col_3": [1, 0, np.nan, 0, 1, np.nan, 1, 0, 1, 1],
            "col_4": ["A", "B", "A", np.nan, "B", np.nan, "C", "A", "B", "C"],
        },
        index=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
    )


@pytest.fixture(scope="function")
def y():
    """
    Fixture.
    """
    return pd.Series([1, 0, 1, 0, 1, 0, 1, 0, 0, 0], index=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10])


@pytest.fixture(scope="function")
def strategies():
    """
    Test strategies.
    """
    return {
        "Simple Median Imputer": SimpleImputer(strategy="median", add_indicator=True),
        "Simple Mean Imputer": SimpleImputer(strategy="mean", add_indicator=True),
        "Iterative Imputer": IterativeImputer(add_indicator=True, n_nearest_features=5, sample_posterior=True),
        "KNN": KNNImputer(n_neighbors=3),
    }


def test_imputation_linear(X, y, strategies, capsys):
    """
    Test imputation linear.
    """
    # Initialize the classifier
    clf = LogisticRegression()
    cmp = ImputationSelector(clf=clf, strategies=strategies, cv=3, model_na_support=False)
    report = cmp.fit_compute(X, y)
    _ = cmp.plot(show=False)

    assert cmp.fitted
    cmp._check_if_fitted()
    assert report.shape[0] == 4

    # Check if there is any prints
    out, _ = capsys.readouterr()
    assert len(out) == 0


def test_imputation_bagging(X, y, strategies, capsys):
    """
    Test bagging.
    """
    # Initialize the classifier
    clf = RandomForestClassifier()
    cmp = ImputationSelector(clf=clf, strategies=strategies, cv=3, model_na_support=False)
    report = cmp.fit_compute(X, y)
    _ = cmp.plot(show=False)

    assert cmp.fitted
    cmp._check_if_fitted()
    assert report.shape[0] == 4

    # Check if there is any prints
    out, _ = capsys.readouterr()
    assert len(out) == 0


@pytest.mark.skipif(os.environ.get("SKIP_LIGHTGBM") == "true", reason="LightGBM tests disabled")
def test_imputation_boosting(X, y, strategies, complex_lightgbm, capsys):
    """
    Test boosting.
    """
    # Initialize the classifier
    clf = complex_lightgbm
    cmp = ImputationSelector(clf=clf, strategies=strategies, cv=3, model_na_support=True)
    report = cmp.fit_compute(X, y)
    _ = cmp.plot(show=False)

    assert cmp.fitted
    cmp._check_if_fitted()
    assert report.shape[0] == 5

    # Check if there is any prints
    out, _ = capsys.readouterr()
    assert len(out) == 0



import numpy as np

from probatus.stat_tests import es, ks


def test_verbosity_true_(capsys):
    """
    Test.
    """
    d1 = np.random.normal(size=1000)
    d2 = d1
    ks(d1, d2, verbose=True)
    captured = capsys.readouterr()
    assert (
        captured.out
        == "\nKS: pvalue = 1.0\n\nKS: Null hypothesis cannot be rejected. Distributions not statistically different.\n"
    )
    es(d1, d2, verbose=True)
    captured = capsys.readouterr()
    assert (
        captured.out
        == "\nES: pvalue = 1.0\n\nES: Null hypothesis cannot be rejected. Distributions not statistically different.\n"
    )


def test_verbosity_false(capsys):
    """
    Test.
    """
    d1 = np.random.normal(size=1000)
    d2 = d1
    ks(d1, d2, verbose=False)
    captured = capsys.readouterr()
    assert captured.out == ""



# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from probatus.binning import AgglomerativeBucketer, QuantileBucketer, SimpleBucketer
from probatus.utils import BaseFitComputePlotClass, preprocess_data, preprocess_labels, shap_to_df


class DependencePlotter(BaseFitComputePlotClass):
    """
    Plotter used to plot SHAP dependence plot together with the target rates.

    Currently it supports tree-based and linear models.

    Args:
        model: classifier for which interpretation is done.

    Example:
    ```python
    from sklearn.datasets import make_classification
    from sklearn.ensemble import RandomForestClassifier
    from probatus.interpret import DependencePlotter

    X, y = make_classification(n_samples=15, n_features=3, n_informative=3, n_redundant=0, random_state=42)
    clf = RandomForestClassifier().fit(X, y)
    bdp = DependencePlotter(clf)
    shap_values = bdp.fit_compute(X, y)

    bdp.plot(feature=2, type_binning='simple')
    ```

    <img src="../img/model_interpret_dep.png"/>
    """

    def __init__(self, clf, verbose=0):
        """
        Initializes the class.

        Args:
            clf (model object):
                Binary classification model or pipeline.

            verbose (int, optional):
                Controls verbosity of the output:

                - 0 - neither prints nor warnings are shown
                - 1 - 50 - only most important warnings regarding data properties are shown (excluding SHAP warnings)
                - 51 - 100 - shows most important warnings, prints of the feature removal process
                - above 100 - presents all prints and all warnings (including SHAP warnings).
        """
        self.clf = clf
        self.verbose = verbose

    def __repr__(self):
        """
        Represent string method.
        """
        return f"Shap dependence plotter for {self.clf.__class__.__name__}"

    def fit(self, X, y, column_names=None, class_names=None, precalc_shap=None, **shap_kwargs):
        """
        Fits the plotter to the model and data by computing the shap values.

        If the shap_values are passed, they do not need to be computed.

        Args:
            X (pd.DataFrame): input variables.

            y (pd.Series): target variable.

            column_names (None, or list of str, optional):
                List of feature names for the dataset. If None, then column names from the X_train dataframe are used.

            class_names (None, or list of str, optional):
                List of class names e.g. ['neg', 'pos']. If none, the default ['Negative Class', 'Positive Class'] are
                used.

            precalc_shap (Optional, None or np.array):
                Precalculated shap values, If provided they don't need to be computed.

            **shap_kwargs:
                keyword arguments passed to
                [shap.Explainer](https://shap.readthedocs.io/en/latest/generated/shap.Explainer.html#shap.Explainer).
                It also enables `approximate` and `check_additivity` parameters, passed while calculating SHAP values.
                The `approximate=True` causes less accurate, but faster SHAP values calculation, while
                `check_additivity=False` disables the additivity check inside SHAP.
        """
        self.X, self.column_names = preprocess_data(X, X_name="X", column_names=column_names, verbose=self.verbose)
        self.y = preprocess_labels(y, y_name="y", index=self.X.index, verbose=self.verbose)

        # Set class names
        self.class_names = class_names
        if self.class_names is None:
            self.class_names = ["Negative Class", "Positive Class"]

        self.shap_vals_df = shap_to_df(self.clf, self.X, precalc_shap=precalc_shap, verbose=self.verbose, **shap_kwargs)

        self.fitted = True
        return self

    def compute(self):
        """
        Computes the report returned to the user, namely the SHAP values generated on the dataset.

        Returns:
            (pd.DataFrame):
                SHAP Values for X.
        """
        self._check_if_fitted()
        return self.shap_vals_df

    def fit_compute(self, X, y, column_names=None, class_names=None, precalc_shap=None, **shap_kwargs):
        """
        Fits the plotter to the model and data by computing the shap values.

        If the shap_values are passed, they do not need to be computed

        Args:
            X (pd.DataFrame):
                Provided dataset.

            y (pd.Series):
                Binary labels for X.

            column_names (None, or list of str, optional):
                List of feature names for the dataset. If None, then column names from the X_train dataframe are used.

            class_names (None, or list of str, optional):
                List of class names e.g. ['neg', 'pos']. If none, the default ['Negative Class', 'Positive Class'] are
                used.

            precalc_shap (Optional, None or np.array):
                Precalculated shap values, If provided they don't need to be computed.

            **shap_kwargs:
                keyword arguments passed to
                [shap.Explainer](https://shap.readthedocs.io/en/latest/generated/shap.Explainer.html#shap.Explainer).
                It also enables `approximate` and `check_additivity` parameters, passed while calculating SHAP values.
                The `approximate=True` causes less accurate, but faster SHAP values calculation, while
                `check_additivity=False` disables the additivity check inside SHAP.

        Returns:
            (pd.DataFrame):
                SHAP Values for X.
        """
        self.fit(X, y, column_names=column_names, class_names=class_names, precalc_shap=precalc_shap, **shap_kwargs)
        return self.compute()

    def plot(
        self,
        feature,
        figsize=(15, 10),
        bins=10,
        type_binning="simple",
        show=True,
        min_q=0,
        max_q=1,
        alpha=1.0,
    ):
        """
        Plots the shap values for data points for a given feature, as well as the target rate and values distribution.

        Args:
            feature (str or int):
                Feature name of the feature to be analyzed.

            figsize ((float, float), optional):
                Tuple specifying size (width, height) of resulting figure in inches.

            bins (int or list[float]):
                Number of bins or boundaries of bins (supplied in list) for target-rate plot.

            type_binning ({'simple', 'agglomerative', 'quantile'}):
                Type of binning to be used in target-rate plot (see :mod:`binning` for more information).

            show (bool, optional):
                If True, the plots are showed to the user, otherwise they are not shown. Not showing plot can be useful,
                when you want to edit the returned axis, before showing it.

            min_q (float, optional):
                Optional minimum quantile from which to consider values, used for plotting under outliers.

            max_q (float, optional):
                Optional maximum quantile until which data points are considered, used for plotting under outliers.

            alpha (float, optional):
                Optional alpha blending value, between 0 (transparent) and 1 (opaque).

        Returns
            (list(matplotlib.axes)):
                List of axes that include the plots.
        """
        self._check_if_fitted()
        if min_q >= max_q:
            raise ValueError("min_q must be smaller than max_q")
        if feature not in self.X.columns:
            raise ValueError("Feature not recognized")
        if type_binning not in ["simple", "agglomerative", "quantile"]:
            raise ValueError("Select one of the following binning methods: 'simple', 'agglomerative', 'quantile'")
        if (alpha < 0) or (alpha > 1):
            raise ValueError("alpha must be a float value between 0 and 1")

        self.min_q, self.max_q, self.alpha = min_q, max_q, alpha

        _ = plt.figure(1, figsize=figsize)
        ax1 = plt.subplot2grid((3, 1), (0, 0), rowspan=2)
        ax2 = plt.subplot2grid((3, 1), (2, 0))

        self._dependence_plot(feature=feature, ax=ax1)
        self._target_rate_plot(feature=feature, bins=bins, type_binning=type_binning, ax=ax2)

        ax2.set_xlim(ax1.get_xlim())

        if show:
            plt.show()
        else:
            plt.close()

        return [ax1, ax2]

    def _dependence_plot(self, feature, ax=None):
        """
        Plots shap values for data points with respect to specified feature.

        Args:
            feature (str or int):
                Feature for which dependence plot is to be created.

            ax (matplotlib.pyplot.axes, optional):
                Optional axis on which to draw plot.

        Returns:
            (matplotlib.pyplot.axes):
                Axes on which plot is drawn.
        """
        if type(feature) is int:
            feature = self.column_names[feature]

        X, y, shap_val = self._get_X_y_shap_with_q_cut(feature=feature)

        ax.scatter(X[y == 0], shap_val[y == 0], label=self.class_names[0], color="lightblue", alpha=self.alpha)

        ax.scatter(X[y == 1], shap_val[y == 1], label=self.class_names[1], color="darkred", alpha=self.alpha)

        ax.set_ylabel("Shap value")
        ax.set_title(f"Dependence plot for {feature} feature")
        ax.legend()

        return ax

    def _target_rate_plot(self, feature, bins=10, type_binning="simple", ax=None):
        """
        Plots the distributions of the specific features, as well as the target rate as function of the feature.

        Args:
            feature (str or int):
                Feature for which to create target rate plot.

            bins (int or list[float]), optional:
                Number of bins or boundaries of desired bins in list.

            type_binning ({'simple', 'agglomerative', 'quantile'}, optional):
                Type of binning strategy used to create bins.

            ax (matplotlib.pyplot.axes, optional):
                Optional axis on which to draw plot.

        Returns:
            (list[float], matplotlib.pyplot.axes, float):
                Tuple of boundaries of bins used, axis on which plot is drawn, total ratio of target (positive over
                negative).
        """
        x, y, shap_val = self._get_X_y_shap_with_q_cut(feature=feature)

        # Create bins if not explicitly supplied
        if type(bins) is int:
            if type_binning == "simple":
                counts, bins = SimpleBucketer.simple_bins(x, bins)
            elif type_binning == "agglomerative":
                counts, bins = AgglomerativeBucketer.agglomerative_clustering_binning(x, bins)
            elif type_binning == "quantile":
                counts, bins = QuantileBucketer.quantile_bins(x, bins)

        # Determine bin for datapoints
        bins[-1] = bins[-1] + 1
        indices = np.digitize(x, bins)
        # Create dataframe with binned data
        dfs = pd.DataFrame({feature: x, "y": y, "bin_index": pd.Series(indices, index=x.index)}).groupby(
            "bin_index", as_index=True
        )

        # Extract target ratio and mean feature value
        target_ratio = dfs["y"].mean()
        x_vals = dfs[feature].mean()

        # Transform the first and last bin to work with plt.hist method
        if bins[0] == -np.inf:
            bins[0] = x.min()
        if bins[-1] == np.inf:
            bins[-1] = x.max()

        # Plot target rate
        ax.hist(x, bins=bins, lw=2, alpha=0.4)
        ax.set_ylabel("Counts")
        ax2 = ax.twinx()
        ax2.plot(x_vals, target_ratio, color="red")
        ax2.set_ylabel("Target rate", color="red", fontsize=12)
        ax2.set_xlim(x.min(), x.max())
        ax.set_xlabel(f"{feature} feature values")

        return bins, ax, target_ratio

    def _get_X_y_shap_with_q_cut(self, feature):
        """
        Extracts all X, y pairs and shap values that fall within defined quantiles of the feature.

        Args:
            feature (str): feature to return values for

        Returns:
            x (pd.Series): selected datapoints
            y (pd.Series): target values of selected datapoints
            shap_val (pd.Series): shap values of selected datapoints
        """
        self._check_if_fitted()
        if feature not in self.X.columns:
            raise ValueError("Feature not found in data")

        # Prepare arrays
        x = self.X[feature]
        y = self.y
        shap_val = self.shap_vals_df[feature]

        # Determine quantile ranges
        x_min = x.quantile(self.min_q)
        x_max = x.quantile(self.max_q)

        # Create filter
        filter = (x >= x_min) & (x <= x_max)

        # Filter and return terms
        return x[filter], y[filter], shap_val[filter]



import numpy as np
import pytest
from sklearn.exceptions import NotFittedError

from probatus.binning import AgglomerativeBucketer, Bucketer, QuantileBucketer, SimpleBucketer, TreeBucketer


@pytest.mark.filterwarnings("ignore:")
def test_deprecations():
    """
    Test.
    """
    x = [1, 2, 1]
    bins = 3
    myBucketer = SimpleBucketer(bin_count=bins)
    myBucketer.fit(x)
    with pytest.deprecated_call():
        myBucketer.counts

    with pytest.deprecated_call():
        myBucketer.boundaries


def test_simple_bins():
    """
    Test.
    """
    x = [1, 2, 1]
    bins = 3
    myBucketer = SimpleBucketer(bin_count=bins)
    with pytest.raises(NotFittedError):
        myBucketer.compute([1, 2])

    myBucketer.fit(x)
    assert len(myBucketer.counts_) == bins
    assert np.array_equal(myBucketer.counts_, np.array([2, 0, 1]))
    assert len(myBucketer.boundaries_) == bins + 1
    np.testing.assert_array_almost_equal(myBucketer.boundaries_, np.array([-np.inf, 1.33333333, 1.66666667, np.inf]))
    # test static method
    counts, boundaries = SimpleBucketer(bin_count=bins).simple_bins(x, bins)
    assert np.array_equal(myBucketer.counts_, counts)
    np.testing.assert_array_almost_equal(myBucketer.boundaries_, boundaries)
    assert repr(myBucketer).startswith("SimpleBucketer")


def test_quantile_bins():
    """
    Test.
    """
    bins = 4
    random_state = np.random.RandomState(0)
    x = random_state.normal(0, 1, size=1000)
    myBucketer = QuantileBucketer(bin_count=bins)
    with pytest.raises(NotFittedError):
        myBucketer.compute([1, 2])
    myBucketer.fit(x)
    assert len(myBucketer.counts_) == bins
    assert np.array_equal(myBucketer.counts_, np.array([250, 250, 250, 250]))
    assert len(myBucketer.boundaries_) == bins + 1
    np.testing.assert_array_almost_equal(
        myBucketer.boundaries_, np.array([-np.inf, -0.7, -0.1, 0.6, np.inf]), decimal=1
    )
    # test static method
    counts, boundaries = QuantileBucketer(bin_count=bins).quantile_bins(x, bins)
    assert np.array_equal(myBucketer.counts_, counts)
    np.testing.assert_array_almost_equal(myBucketer.boundaries_, boundaries)
    # test inf edges
    counts, boundaries = QuantileBucketer(bin_count=bins).quantile_bins(x, bins, inf_edges=True)
    assert boundaries[0] == -np.inf
    assert boundaries[-1] == np.inf
    assert repr(myBucketer).startswith("QuantileBucketer")


def test_agglomerative_clustering_new():
    """
    Test.
    """

    x = [0.5, 1, 1, 1, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4.5]
    bins = 4
    myBucketer = AgglomerativeBucketer(bin_count=bins)
    with pytest.raises(NotFittedError):
        myBucketer.compute([1, 2])
    myBucketer.fit(x)
    assert len(myBucketer.counts_) == bins
    print(myBucketer.counts_)
    assert np.array_equal(myBucketer.counts_, np.array([4, 2, 5, 3]))
    assert len(myBucketer.boundaries_) == bins + 1
    np.testing.assert_array_almost_equal(myBucketer.boundaries_, np.array([-np.inf, 1.5, 2.5, 3.5, np.inf]), decimal=2)
    # test static method
    counts, boundaries = AgglomerativeBucketer(bin_count=bins).agglomerative_clustering_binning(x, bins)
    assert np.array_equal(myBucketer.counts_, counts)
    np.testing.assert_array_almost_equal(myBucketer.boundaries_, boundaries)
    assert repr(myBucketer).startswith("AgglomerativeBucketer")


def test_compute():
    """
    Test.
    """
    x = np.arange(10)
    bins = 5
    myBucketer = QuantileBucketer(bins)
    x_new = x
    with pytest.raises(NotFittedError):
        assert myBucketer.compute(x_new)
    myBucketer.fit(x)
    assert len(myBucketer.compute(x_new)) == bins
    np.testing.assert_array_equal(myBucketer.counts_, myBucketer.compute(x_new))
    np.testing.assert_array_equal(myBucketer.counts_, myBucketer.fit_compute(x_new))
    x_new = x + 100
    np.testing.assert_array_equal(np.array([0, 0, 0, 0, 10]), myBucketer.compute(x_new))
    x_new = x - 100
    np.testing.assert_array_equal(np.array([10, 0, 0, 0, 0]), myBucketer.compute(x_new))
    x_new = [1, 1, 1, 4, 4, 7]
    np.testing.assert_array_equal(np.array([3, 0, 2, 1, 0]), myBucketer.compute(x_new))


def test_quantile_with_unique_values():
    """
    Test.
    """
    np.random.seed(42)
    dist_0_1 = np.random.uniform(size=20)
    dist_peak_at_0 = np.zeros(shape=20)

    skewed_dist = np.hstack((dist_0_1, dist_peak_at_0))
    actual_out = QuantileBucketer(10).quantile_bins(skewed_dist, 10)

    expected_out = (
        np.array([20, 4, 4, 4, 4, 4]),
        np.array([0.0, 0.01894458, 0.23632033, 0.42214475, 0.60977678, 0.67440958, 0.99940487]),
    )

    assert (actual_out[0] == expected_out[0]).all()


def test_tree_bucketer():
    """
    Test.
    """
    x = np.array(
        [
            0.0,
            0.2,
            0.4,
            0.6,
            0.8,
            1.0,
            1.2,
            1.4,
            1.6,
            1.8,
            2.0,
            2.2,
            2.4,
            2.6,
            2.8,
            3.0,
            3.2,
            3.4,
            3.6,
            3.8,
            4.0,
            4.2,
            4.4,
            4.6,
            4.8,
            5.0,
            5.2,
            5.4,
            5.6,
            5.8,
            6.0,
            6.2,
            6.4,
            6.6,
            6.8,
            7.0,
            7.2,
            7.4,
            7.6,
            7.8,
            8.0,
            8.2,
            8.4,
            8.6,
            8.8,
            9.0,
            9.2,
            9.4,
            9.6,
            9.8,
        ]
    )

    y = np.array(
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            1,
            0,
            0,
            1,
            0,
            1,
            1,
            0,
            0,
            0,
            1,
            1,
            0,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
        ]
    )

    myTreeBucketer = TreeBucketer(inf_edges=False, max_depth=3, min_samples_leaf=10, random_state=42)

    with pytest.raises(NotFittedError):
        myTreeBucketer.compute([1, 2])

    myTreeBucketer.fit(x, y)

    assert all(myTreeBucketer.counts_ == np.array([21, 15, 14]))
    assert myTreeBucketer.bin_count == 3
    assert all(myTreeBucketer.boundaries_ - np.array([0.0, 4.1, 7.1, 9.8]) < 0.01)

    # If infinite edges is False, it must get the edges of the x array
    assert myTreeBucketer.boundaries_[0] == 0
    assert myTreeBucketer.boundaries_[-1] == 9.8

    myTreeBucketer = TreeBucketer(inf_edges=True, max_depth=3, min_samples_leaf=10, random_state=42)

    myTreeBucketer.fit(x, y)
    # check that the infinite edges is True, then edges must be infinite
    assert myTreeBucketer.boundaries_[0] == -np.inf
    assert myTreeBucketer.boundaries_[-1] == +np.inf


def test_tree_bucketer_dependence():
    """
    Test.
    """
    x = np.arange(0, 10, 0.01)
    y = [1 if z < 0.5 else 0 for z in np.random.uniform(size=x.shape[0])]

    # Test number of leaves is always within the expected ranges
    myTreeBucketer = TreeBucketer(inf_edges=False, max_depth=3, min_samples_leaf=10, random_state=42).fit(x, y)
    assert myTreeBucketer.bin_count <= np.power(2, myTreeBucketer.tree.max_depth)

    # Test number of leaves is always within the expected ranges
    myTreeBucketer = TreeBucketer(inf_edges=False, max_depth=6, min_samples_leaf=1, random_state=42).fit(x, y)
    assert myTreeBucketer.bin_count <= np.power(2, myTreeBucketer.tree.max_depth)

    # Test that the counts per bin never drop below min_samples_leaf
    myTreeBucketer = TreeBucketer(inf_edges=False, max_depth=6, min_samples_leaf=100, random_state=42).fit(x, y)
    assert all([x >= myTreeBucketer.tree.min_samples_leaf for x in myTreeBucketer.counts_])

    myTreeBucketer = TreeBucketer(inf_edges=False, max_depth=6, min_samples_leaf=200, random_state=42).fit(x, y)
    assert all([x >= myTreeBucketer.tree.min_samples_leaf for x in myTreeBucketer.counts_])

    # Test that if the leaf is set to the number of entries,it raises an Error
    myTreeBucketer = TreeBucketer(inf_edges=False, max_depth=6, min_samples_leaf=x.shape[0], random_state=42)

    with pytest.raises(ValueError):
        assert myTreeBucketer.fit(x, y)

    # Test that if the leaf is set to the number of entries-1, it returns only one bin
    myTreeBucketer = TreeBucketer(inf_edges=False, max_depth=6, min_samples_leaf=x.shape[0] - 1, random_state=42).fit(
        x, y
    )
    assert myTreeBucketer.bin_count == 1
    assert all([x >= myTreeBucketer.tree.min_samples_leaf for x in myTreeBucketer.counts_])


def test_tree_binning():
    """
    Test binning with a decisiontree.
    """
    x = [1, 2, 2, 5, 3]
    y = [0, 0, 1, 1, 1]
    myBucketer = TreeBucketer(inf_edges=True, max_depth=2, min_impurity_decrease=0.001)
    myBucketer.fit(x, y)
    assert myBucketer.boundaries_ == [-np.inf, 1.5, 2.5, np.inf]
    assert myBucketer.bin_count == 3
    assert myBucketer.counts_ == [1, 2, 2]

    myBucketer = TreeBucketer(max_depth=2, min_impurity_decrease=0.001)
    myBucketer.fit(x, y)
    assert myBucketer.boundaries_ == [1, 1.5, 2.5, 5]
    assert myBucketer.bin_count == 3
    assert myBucketer.counts_ == [1, 2, 2]


def test_compute_counts_per_bin():
    """
    Test for checking if counts per bin are correctly computed.
    """
    x = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    boundaries = [0, 1, 5, 6, 10, 11]  # down boundary < current value <= up boundary
    np.testing.assert_array_almost_equal(Bucketer._compute_counts_per_bin(x, boundaries), np.array([1, 4, 1, 4, 0]))
